from typing import Union, NoReturn

from calendar_view.config import style
from calendar_view.core.config import CalendarConfig
from calendar_view.core.time_utils import *


class EventStyle(object):
    """
    Defined the style of the painted event.
    """
    def __init__(self, event_border: Tuple[int, int, int, int] = None, event_fill: Tuple[int, int, int, int] = None) -> None:
        """
        :param event_border: the color of the border as a tuple (r, g, b, a)
            Example: (120, 180, 120, 240) - green
        :param event_fill: the color of the background as a tuple (r, g, b, a)
            Example: (196, 234, 188, 180) - light green
        """
        self.event_border = event_border if event_border else style.event_border_default
        self.event_fill = event_fill if event_fill else style.event_fill_default

    def __repr__(self) -> str:
        return f'EventStyle[event_border: {self.event_border}, event_fill: {self.event_fill}]'

    def __eq__(self, other) -> bool:
        return isinstance(other, EventStyle) \
               and self.event_border == other.event_border \
               and self.event_fill == other.event_fill


class EventStyles(object):
    """
    Predefined colors
    """
    GREEN = EventStyle(event_border=(120, 180, 120, 240), event_fill=(196, 234, 188, 180))
    RED = EventStyle(event_border=(220, 50, 50, 240), event_fill=(220, 50, 50, 180))
    BLUE = EventStyle(event_border=(100, 100, 220, 240), event_fill=(150, 150, 234, 180))
    GRAY = EventStyle(event_border=(110, 110, 110, 240), event_fill=(200, 200, 200, 190))


class Event(object):
    def __init__(self, title: str = None, notes: str = None, day_of_week: int = None,
                 day: Union[date, datetime, str] = None, start: Union[datetime, time, str] = None,
                 end: Union[datetime, time, str] = None, style: EventStyle = None) -> None:
        if day_of_week and not (0 <= day_of_week <= 6):
            raise ValueError("'day_of_week' has to be in the interval [0, 6]. Current value is: {}".format(day_of_week))
        if not start:
            raise ValueError("'start' argument is required")
        if not end:
            raise ValueError("'end' argument is required")

        self.title: Optional[str] = title
        self.notes: Optional[str] = notes
        self.style: EventStyle = style if style else EventStyle()

        # parse date of the event
        self.__start_date: date = self.__parse_start_date(day_of_week, day, start)
        self.__end_date: date = self.__parse_end_date(day_of_week, day, end)
        if self.__end_date is None:
            self.__end_date: date = self.__start_date
        self.__day_of_week: Optional[int] = None
        if day_of_week is not None:
            self.__day_of_week = day_of_week

        # parse the time of the start and end of the event
        self.__start_time: time = Event.__parse_time(start)
        self.__end_time: time = Event.__parse_time(end)

        self.cascade_total: int = 1
        self.cascade_index: int = 1
        self.cascade_group: int = 0
        # run additional validation
        self.__validate()


    @staticmethod
    def __parse_start_date(day_of_week: Optional[int],
                           day: Union[date, datetime, str],
                           start: Optional[Union[datetime, time, str]]) -> Optional[date]:
        if day_of_week is not None and day is not None:
            raise ValueError("'day_of_week' is defined together with a 'day'.")
        if day_of_week is not None and (start is not None and isinstance(start, datetime)):
            raise ValueError("'day_of_week' is defined together with a 'start' as a datetime.")
        return Event.__parse_date(day_of_week, day, start)

    @staticmethod
    def __parse_end_date(day_of_week: Optional[int],
                         day: Optional[Union[date, datetime, str]],
                         end: Optional[Union[datetime, time, str]]) -> Optional[date]:
        if day_of_week is not None and day is not None:
            raise ValueError("'day_of_week' is defined together with a 'day'.")
        if day_of_week is not None and (end is not None and isinstance(end, datetime)):
            raise ValueError("'day_of_week' is defined together with a 'end' as a datetime.")
        return Event.__parse_date(day_of_week, day, end)

    @staticmethod
    def __parse_date(day_of_week: Optional[int],
                     day: Optional[Union[date, datetime, str]],
                     time_entry: Optional[Union[datetime, time, str]]) -> Optional[date]:
        if day_of_week:
            return None
        if day:
            if isinstance(day, date):
                return day
            if isinstance(day, datetime):
                return day.date()
            if isinstance(day, str):
                return parse_date(day)
        if time_entry and isinstance(time_entry, datetime):
            return time_entry.date()
        return None

    @staticmethod
    def __parse_time(time_entry: Union[datetime, time, str]) -> Optional[time]:
        if isinstance(time_entry, datetime):
            return time_entry.time()
        if isinstance(time_entry, time):
            return time_entry
        if isinstance(time_entry, str):
            return parse_time(time_entry)

    def __validate(self):
        if self.__start_date is None and self.__day_of_week is None:
            raise ValueError("Either date of the start event or the day of the week has to be defined.")
        if self.__end_date is None and self.__day_of_week is None:
            raise ValueError("Either date of the end event or the day of the week has to be defined.")
        if self.__start_date and self.__end_date and self.__end_date < self.__start_date:
            raise ValueError("Event's start date has to be before end date")
        if self.__start_time == self.__end_time and self.__end_time < self.__start_time:
            raise ValueError("Event's start time has to be before end time")

    def get_start_date(self, config: CalendarConfig) -> date:
        if self.__start_date:
            return self.__start_date
        if config is None:
            return current_week_day(self.__day_of_week)
        return self.__get_date_by_day_of_week(config)

    def get_end_date(self, config: CalendarConfig) -> date:
        if self.__end_date:
            return self.__end_date
        if config is None:
            return current_week_day(self.__day_of_week)
        return self.__get_date_by_day_of_week(config)

    def __get_date_by_day_of_week(self, config: CalendarConfig) -> date:
        if not config:
            raise ValueError("'config' must be defined")

        start_date, end_date = config.get_date_range()
        days_duration: int = (end_date - start_date).days
        if days_duration > 7:
            raise ValueError(f"Cannot use 'day_of_week' parameter for the period more than a week. "
                             f"Your range is [{start_date}, {end_date}]")

        result_date = week_day_for_date(start_date, self.__day_of_week)
        if start_date <= result_date or end_date <= result_date + timedelta(weeks=1):
            return result_date
        return result_date + timedelta(weeks=1)

    def get_duration_seconds(self, config: CalendarConfig) -> float:
        duration: timedelta = datetime.combine(self.get_end_date(config), self.end_time) - datetime.combine(self.get_start_date(config), self.start_time)
        return duration.total_seconds()

    @property
    def start_time(self) -> time:
        return self.__start_time

    @property
    def end_time(self) -> time:
        return self.__end_time

    def __repr__(self) -> str:
        return f'Event[title: {self.title}, notes: {self.notes}, style: {self.style}, ' \
               f'day_of_week: {self.__day_of_week}, start_date: {self.__start_date}, end_date: {self.__end_date}, ' \
               f'start_time: {self.start_time}, end_time: {self.end_time},' \
               f'cascade_group: {self.cascade_group}, cascade_total: {self.cascade_total}, ' \
               f'cascade_index: {self.cascade_index}]'

    def __eq__(self, other) -> bool:
        return isinstance(other, Event) \
               and self.title == other.title \
               and self.notes == other.notes \
               and self.style == other.style \
               and self.__day_of_week == other.__day_of_week \
               and self.__start_date == other.__start_date \
               and self.__end_date == other.__end_date \
               and self.start_time == other.start_time \
               and self.end_time == other.end_time \
               and self.cascade_group == other.cascade_group \
               and self.cascade_total == other.cascade_total \
               and self.cascade_index == other.cascade_index



import logging
from datetime import date, timedelta
from typing import Tuple, List, Literal

from calendar_view.core import time_utils
from calendar_view.core.utils import StringUtils

VerticalAlign = Literal['top', 'center', 'bottom']


class CalendarConfig(object):
    """
    Mode will override some parameters.
    Available modes:
    'auto' - modes 'week' + 'day_hours'
    'week' - show current week
    'day_hours' - show hours range '8:00 - 22:00'
    'working_hours' - show hours range '8:00 - 19:00'
    """
    def __init__(self,
                 lang: str = 'en',
                 title: str = '',
                 dates: str = None,
                 days: int = 7,
                 hours: str = None,
                 mode: str = None,
                 show_date: bool = True,
                 show_year: bool = False,
                 legend: bool = None,
                 title_vertical_align: VerticalAlign = 'center'):
        self.lang = lang
        self.title = title
        self.dates = dates
        self.days = days
        self.hours = hours
        self.mode = mode
        self.show_date = show_date
        self.show_year = show_year
        self.legend = legend
        self.title_vertical_align = title_vertical_align
        self._configure_mode()

    def _configure_mode(self):
        if self.mode is None:
            return
        modes: List[str] = self.mode.replace(' ', '').split(',')
        if 'auto' in modes:
            modes.append('week')
            modes.append('day_hours')
        if 'week' in modes:
            self.dates = '{} - {}'.format(
                time_utils.current_week_day(0).strftime('%Y-%m-%d'),
                time_utils.current_week_day(6).strftime('%Y-%m-%d')
            )
        if 'day_hours' in modes:
            self.hours = '8:00 - 22:00'
        if 'working_hours' in modes:
            self.hours = '8:00 - 19:00'

    def validate(self):
        if StringUtils.is_blank(self.lang):
            raise Exception("Parameter 'lang' is empty. Language has to be specified")
        if not (0 < self.days <= 14):
            raise Exception("Parameter 'days' can be in interval [1, 14]")
        self.get_hours_range()
        self.get_date_range()
        if self.dates and self.days:
            logging.warning("Both parameters 'days' and 'dates' are used. 'days' value will be skipped.")
        if self.show_year and not self.show_date:
            logging.warning("'show_year' is set to True, but date wont be displayed, because 'show_date' is False.")

    def get_date_range(self) -> Tuple[date, date]:
        """
        Returns tuple of start and end day for visualisation. For example: 'date(2019, 05, 17), date(2019, 05, 20)'
        """
        if StringUtils.is_not_blank(self.dates):
            return time_utils.parse_date_interval(self.dates)
        if self.days:
            return date.today(), date.today() + timedelta(days=self.days - 1)

        logging.warning("Date range is not defined. Using default range 'Mo - Su'.")
        return time_utils.current_week_day(0), time_utils.current_week_day(6)

    def get_hours_range(self) -> Tuple[int, int]:
        """
        Returns tuple of start and end hour for visualisation. For example: '10, 18'
        """
        if StringUtils.is_blank(self.hours):
            return 0, 24

        start, end = time_utils.parse_time_interval(self.hours)
        fixed_end = end.hour if end.minute == 0 else end.hour + 1
        if fixed_end == 0:
            fixed_end = 24

        return start.hour, fixed_end



from collections import defaultdict
from typing import List, NoReturn

from PIL import Image, ImageDraw

from calendar_view.config import style
from calendar_view.core.calendar_events import CalendarEvents
from calendar_view.core.calendar_grid import CalendarGrid
from calendar_view.core.config import CalendarConfig
from calendar_view.core.event import Event
from calendar_view.core.utils import StringUtils


class Calendar:
    @staticmethod
    def build(config: CalendarConfig = CalendarConfig()):
        cal = Calendar(config)
        cal.draw_grid()
        return cal

    def __init__(self, config: CalendarConfig):
        self.config = config
        self.grid = CalendarGrid(config)
        self.events = CalendarEvents(config)
        self.full_image: Image = None

    def draw_grid(self):
        self.grid.draw_grid()
        self.events.draw_grid(self.grid.get_size())

    def add_events(self, events: List[Event]) -> None:
        """
        Adds the input events to the list to draw them later.
        :param events: the list of events
        """
        for e in events:
            self.events.add_event(e)

    def add_event(self, *events: Event, **kwargs) -> None:
        """
        Adds the event(s) to the list to draw them later.
        :param events: the event objects
        :param kwargs: the input arguments for the Event constructor
        """
        if events:
            for event in events:
                self.events.add_event(event)
        if kwargs:
            self.events.add_event(Event(**kwargs))

    def save(self, filename: str) -> NoReturn:
        self.events.group_cascade_events()
        self._build_image()
        self.full_image.save(filename, "PNG")

    def _build_image(self):
        grid_image: Image = self.grid.get_image()
        event_image: Image = self.events.draw_events()
        legend: Image = self.events.draw_legend()

        events: Image = Image.alpha_composite(grid_image, event_image)
        combined: Image = self._combine_image(events, self.config.title, legend)

        self.full_image = Image.new("RGBA", combined.size, style.image_bg)
        self.full_image = Image.alpha_composite(self.full_image, combined)

    def destroy(self):
        self.grid.destroy()
        self.events.destroy()
        del self.full_image

    @staticmethod
    def _combine_image(events: Image, title: str, legend: Image):
        """
        Add title and combine all images into one.
        """
        if StringUtils.is_blank(title) and legend is None:
            return events

        event_width, event_height = events.size
        legend_width, legend_height = (0, 0) if legend is None else legend.size
        title_size = style.title_font.getsize_multiline(title)
        title_width = title_size[0] + style.title_padding_left + style.title_padding_right
        title_height = title_size[1] + style.title_padding_top + style.title_padding_bottom
        final_width = max(event_width, title_width, legend_width)

        combined: Image = Image.new("RGBA", (final_width, event_height + title_height + legend_height), (0, 0, 0, 0))
        draw = ImageDraw.Draw(combined)
        # title
        title_padding_left = max(style.title_padding_left, (final_width - title_size[0]) / 2)
        draw.multiline_text((title_padding_left, style.title_padding_top), title, align='center',
                            font=style.title_font, fill=style.title_color)
        # events
        events_start = (int((final_width - event_width) / 2), title_height)
        combined.paste(events, events_start)
        if legend is not None:
            legend_start = (0, title_height + event_height)
            combined.paste(legend, legend_start)

        return combined



day_of_week_i18n = {
    'en': [
        'Mo',
        'Tu',
        'We',
        'Th',
        'Fr',
        'Sa',
        'Su'
    ],
    'ru': [
        'Пн',
        'Вт',
        'Ср',
        'Чт',
        'Пт',
        'Сб',
        'Вс'
    ],
    'ua': [
        'Пн',
        'Вт',
        'Ср',
        'Чт',
        'Пт',
        'Сб',
        'Нд'
    ],
}


def days_of_week(lang: str = 'en'):
    """
    Returns the localization for the names of all days of the week
    """
    if lang not in day_of_week_i18n.keys():
        raise ValueError('Not defined for language ' + lang)
    return day_of_week_i18n[lang]


def day_of_week(day_of_week: int, lang: str = 'en'):
    """
    Returns the localization for the name of particular day of the week
    :param day_of_week: 0 - Monday, 1 - Tuesday, ..., 6 - Sunday
    """
    if lang not in day_of_week_i18n.keys():
        raise ValueError('Not defined for language ' + lang)
    if not (0 <= day_of_week <= 6):
        raise ValueError('Day of week has to be in range [0, 6]. Currect value: ' + str(day_of_week))
    return day_of_week_i18n[lang][day_of_week]



from PIL.ImageDraw import ImageDraw


def draw_rounded_rectangle(draw: ImageDraw, xy, corner_radius, fill=None, outline=None, width=None):
    rad = corner_radius
    upper_left = xy[0]
    bottom_right = xy[1]
    draw.rectangle(
        [
            (upper_left[0], upper_left[1] + rad),
            (bottom_right[0], bottom_right[1] - rad)
        ],
        fill=fill,
    )
    draw.rectangle(
        [
            (upper_left[0] + rad, upper_left[1]),
            (bottom_right[0] - rad, bottom_right[1])
        ],
        fill=fill,
    )
    draw.pieslice([upper_left, (upper_left[0] + rad * 2, upper_left[1] + rad * 2)],
        180, 270, fill=fill
    )
    draw.pieslice([(bottom_right[0] - rad * 2, bottom_right[1] - rad * 2), bottom_right],
        0, 90, fill=fill
    )
    draw.pieslice([(upper_left[0], bottom_right[1] - rad * 2), (upper_left[0] + rad * 2, bottom_right[1])],
        90, 180, fill=fill
    )
    draw.pieslice([(bottom_right[0] - rad * 2, upper_left[1]), (bottom_right[0], upper_left[1] + rad * 2)],
        270, 360, fill=fill
    )

    border_delta = width/2 - 1
    if width > 0:
        draw.line([(upper_left[0] + rad, upper_left[1] + border_delta), (bottom_right[0] - rad, upper_left[1] + border_delta)],
                  fill=outline, width=width)
        draw.line([(bottom_right[0] - border_delta, upper_left[1] + rad), (bottom_right[0] - border_delta, bottom_right[1] - rad)],
                  fill=outline, width=width)
        draw.line([(bottom_right[0] - rad, bottom_right[1] - border_delta), (upper_left[0] + rad, bottom_right[1] - border_delta)],
                  fill=outline, width=width)
        draw.line([(upper_left[0] + border_delta, bottom_right[1] - rad), (upper_left[0] + border_delta, upper_left[1] + rad)],
                  fill=outline, width=width)

        draw.arc([upper_left, (upper_left[0] + rad * 2, upper_left[1] + rad * 2)],
                 180, 270, fill=outline, width=width)
        draw.arc([(bottom_right[0] - rad * 2, upper_left[1]), (bottom_right[0], upper_left[1] + rad * 2)],
                 270, 360, fill=outline, width=width)
        draw.arc([(bottom_right[0] - rad * 2, bottom_right[1] - rad * 2), bottom_right],
                 0, 90, fill=outline, width=width)
        draw.arc([(upper_left[0], bottom_right[1] - rad * 2), (upper_left[0] + rad * 2, bottom_right[1])],
                 90, 180, fill=outline, width=width)



import logging
from datetime import time

from calendar_view.core.config import CalendarConfig
from calendar_view.core.event import Event


class InputData(object):
    def __init__(self, config: CalendarConfig, events: list):
        self.config = config
        self.events = events


def validate_config(config: CalendarConfig):
    config.validate()


def validate_event(event: Event, config: CalendarConfig):
    start_date, end_date = config.get_date_range()
    start_time = time(hour=config.get_hours_range()[0])
    end_hour = config.get_hours_range()[1]
    if not (start_date <= event.get_start_date(config) <= end_date):
        logging.warning("Event can't be shown, because it is not in configured date range: {} not in [{}, {}]".format(
            event.get_start_date(config).strftime('%Y-%m-%d'),
            start_date.strftime('%Y-%m-%d'),
            end_date.strftime('%Y-%m-%d')
        ))
    if event.start_time < start_time:
        logging.warning("Event can't be shown, because its start is before time range: {} is before {}".format(
            event.start_time.strftime('%H:%M'),
            start_time.strftime('%H:%M')
        ))
    if end_hour < 24 and time(hour=end_hour) < event.end_time:
        logging.warning("Event can't be shown, because its end is after time range: {} is before {}".format(
            event.end_time.strftime('%H:%M'),
            time(hour=end_hour).strftime('%H:%M')
        ))


def validate_events(events: list, config: CalendarConfig):
    for e in events:
        validate_event(e, config)


def validate_data(data: InputData):
    validate_config(data.config)
    validate_events(data.events, data.config)



from datetime import date, timedelta
from typing import Tuple

from PIL import Image, ImageDraw

from calendar_view.config import i18n, style
from calendar_view.core.config import CalendarConfig


class CalendarGrid(object):
    def __init__(self, config: CalendarConfig):
        self.config = config
        self._grid_image: Image = None
        self._grid_draw: ImageDraw = None

    def get_image(self) -> Image:
        return self._grid_image

    def get_size(self) -> Tuple[float, float]:
        return self._grid_image.size

    def draw_grid(self):
        date_from, date_to = self.config.get_date_range()
        day_count = (date_to - date_from).days + 1
        hour_from, hour_to = self.config.get_hours_range()
        hour_count = hour_to - hour_from
        day_height = hour_count * style.hour_height
        size = (day_count * style.day_width + 2 * style.padding_horizontal, style.hour_height + day_height + 2 * style.padding_vertical)
        self._grid_image = Image.new("RGBA", size, (0, 0, 0, 0))
        self._grid_draw = ImageDraw.Draw(self._grid_image)

        # draw hours
        table_width = day_count * style.day_width
        x = (style.padding_horizontal, style.padding_horizontal + table_width)
        for i in range(1, hour_count + 2):
            y = style.padding_vertical + i * style.hour_height
            self._grid_draw.line([(x[0], y), (x[1], y)], fill=style.line_hour_color, width=style.line_hour_width)

        # draw days
        for i in range(day_count + 1):
            x = self.__get_event_x(i)
            y = style.padding_vertical + style.hour_height
            self._grid_draw.line([(x[0], y), (x[0], y + day_height)], fill=style.line_day_color, width=style.line_day_width)

        # write hour numbers
        for i in range(hour_count + 1):
            text = str(hour_from + i)
            text_size = style.hour_number_font.getsize(text)
            x = style.padding_horizontal - text_size[0] - 10
            y = style.padding_vertical + style.hour_height + i * style.hour_height - text_size[1] / 2

            self._grid_draw.text((x, y), text, font=style.hour_number_font, fill=style.hour_number_color)

        # write day of week
        for i in range(day_count):
            day = date_from + timedelta(days=i)
            text = self._get_day_title(day)
            text_size = style.day_of_week_font.getsize(text)
            x = style.padding_horizontal + i * style.day_width + style.day_width / 2 - text_size[0] / 2
            y = style.padding_vertical + text_size[1] / 2
            self._grid_draw.text((x, y), text, font=style.day_of_week_font, fill=style.day_of_week_color)

    def destroy(self):
        del self._grid_image
        del self._grid_draw

    def _get_day_title(self, day: date) -> str:
        weekday = i18n.day_of_week(day.weekday(), self.config.lang)
        if not self.config.show_date:
            return weekday
        else:
            date_value = day.strftime('%d.%m') + (day.strftime('.%Y') if self.config.show_year else '')
            return '{}, {}'.format(weekday, date_value)

    @staticmethod
    def __get_event_x(day_number: int):
        x_start = style.padding_horizontal + day_number * style.day_width
        return x_start, x_start + style.day_width



import setuptools

with open("README.rst", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="calendar-view",
    version="2.1.1",
    author="Oleksandr Sakhnevych, Daniil Limariev",
    author_email="o.sakhnevych@gmail.com, danillim02102003@gmail.com",
    description="Library provides a graphical view of the calendar.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/sakhnevych/CalendarView",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'Pillow',
    ],
    python_requires='>=3.6',
)



from PIL import ImageFont
from pkg_resources import resource_filename

font_path: str = 'Roboto-Regular.ttf'


def image_font(size: int):
    path: str = resource_filename('calendar_view.resources.fonts', font_path)
    return ImageFont.truetype(path, size)


image_bg = (255, 255, 255, 255)

hour_height = 50
day_width = 400
padding_horizontal = 60
padding_vertical = 30

title_font = image_font(50)
title_color = 'black'
title_padding_left = 30
title_padding_right = 30
title_padding_top = 30
title_padding_bottom = 20

hour_number_font = image_font(22)
hour_number_color = 'black'

day_of_week_font = image_font(28)
day_of_week_color = 'black'

line_day_color = (150, 150, 150, 255)
line_day_width = 5
line_hour_color = (180, 180, 180, 210)
line_hour_width = 2

event_border_width = 4
event_radius = 14
event_border_default = (120, 180, 120, 240)
event_fill_default = (196, 234, 188, 210)

event_title_font = image_font(36)
event_title_color = 'black'
event_notes_font = image_font(26)
event_notes_color = 'gray'
event_padding: int = 20
event_title_margin: int = 20

legend_spacing = 20
legend_padding_top = 40
legend_padding_bottom = 70
legend_padding_left = 70
legend_padding_right = 40
legend_name_font = image_font(28)
legend_name_color = 'black'

# https://stackoverflow.com/questions/7510313/transparent-png-in-pil-turns-out-not-to-be-transparent



import re
from datetime import time, date, timedelta, datetime
from typing import Tuple, Optional

from calendar_view.core.utils import StringUtils

MAX_DAYS_RANGE_ALLOWED = 15
ZERO_TIME = time(0, 0)
weekday_dict = {
    'mo': 0,
    'tu': 1,
    'we': 2,
    'th': 3,
    'fr': 4,
    'sa': 5,
    'su': 6,
    'пн': 0,
    'вт': 1,
    'ср': 2,
    'чт': 3,
    'пт': 4,
    'сб': 5,
    'вс': 6,
    'нд': 6,
}
time_regex = re.compile(r'^([0-9]{1,2})(:([0-9]{2}))?$')


def parse_time(value: str) -> Optional[time]:
    """
    Parses time value. Can be used format: 'hh:mm' or 'hh'
    """
    if StringUtils.is_blank(value):
        return None
    m = time_regex.search(value)
    if m is None:
        raise ValueError("Wrong time format: {}. Use: 'hh:mm' or 'hh'".format(value))
    hour = int(m.group(1))
    minute_match = m.group(3)
    minute = int(minute_match) if minute_match else 0
    if hour == 24 and minute == 0:
        hour = 0
    if (0 <= hour <= 23) and (0 <= minute <= 59):
        return time(hour=hour, minute=minute)
    else:
        raise ValueError("Wrong time format: {}. Use: 'hh:mm' or 'hh'".format(value))


def is_valid_time(value: str) -> bool:
    try:
        return parse_time(value) is not None
    except ValueError:
        return False


def parse_time_interval(interval: str) -> Optional[Tuple[time, time]]:
    """
    Important: If both start and end date is 00:00 or 24:00, the interval will be the full day 924 hours)
    """
    interval = interval.strip()
    if interval is None:
        return None
    if '-' not in interval:
        raise ValueError('Wrong interval format: {}. Use: \'hh:mm - hh:mm\''.format(interval))

    start_value, end_value = interval.split('-')
    start = parse_time(start_value.strip())
    end = parse_time(end_value.strip())

    if end <= start and end != ZERO_TIME:
        raise ValueError('Start time has to be before end time: {} - {}'.format(start, end))

    return start, end


def week_day_for_date(d, weekday) -> date:
    days_ahead = weekday - d.weekday()
    return d + timedelta(days_ahead)


def current_week_day(weekday) -> date:
    return week_day_for_date(date.today(), weekday)


def parse_date(value: str) -> Optional[date]:
    """
    Allowed formats:
    'YYYY.mm.dd'
    'dd.mm.YYYY'
    'dd.mm.YY'
    'dd.mm' - current year

    Both '-', '.' and '/' can be used as delimiter.

    Allowed year range: [1900, 2100]
    Specification: ISO 8601
    """
    weekday = weekday_dict.get(value.lower())
    if weekday is not None:
        return current_week_day(weekday)

    dash = '-' in value
    dot = '.' in value
    slash = '/' in value
    sum = int(dot) + int(slash) + int(dash)
    if sum == 0:
        raise ValueError('Wrong date format: {}'.format(value))
    elif sum > 1:
        raise ValueError("Wrong date format: {}. Use only one type of separator: '-', '.' or '/'".format(value))

    parsed = value.split('-' if dash else '.' if dot else '/')
    if not (2 <= len(parsed) <= 3):
        raise ValueError('Wrong date format: {}.')

    result = _parse_date(parsed)
    if not (1900 <= result.year <= 2100):
        raise ValueError('Wrong date range: {}. Can use only dates with year from 1900 to 2100.')
    return result


def _parse_date(parsed: list) -> date:
    try:
        # format 'dd.mm'
        if len(parsed) == 2:
            return date(year=date.today().year, month=int(parsed[1]), day=int(parsed[0]))

        # format 'YYYY.mm.dd'
        if len(parsed[0]) == 4:
            return date(year=int(parsed[0]), month=int(parsed[1]), day=int(parsed[2]))

        # format 'dd.mm.YYYY'
        if len(parsed[2]) == 4:
            return date(year=int(parsed[2]), month=int(parsed[1]), day=int(parsed[0]))

        # format 'dd.mm.YY'
        if len(parsed[2]) == 2:
            return date(year=int(parsed[2]), month=int(parsed[1]), day=int(parsed[0]))
    except ValueError:
        raise ValueError('Wrong date format: {}.')


def parse_date_interval(dates: str) -> Optional[Tuple[date, date]]:
    dates = dates.strip()
    if StringUtils.is_blank(dates):
        return None

    if ' - ' not in dates:
        one_day: date = parse_date(dates.strip())
        return one_day, one_day

    start_value, end_value = dates.split(' - ')
    start: date = parse_date(start_value.strip())
    end: date = parse_date(end_value.strip())

    if start_value.strip().lower() in weekday_dict and end_value.strip().lower() in weekday_dict and end <= start:
        end = end + timedelta(weeks=1)

    if end <= start:
        raise ValueError('Start date has to be before end date: {} - {}'.format(start, end))
    if (end - start).days > MAX_DAYS_RANGE_ALLOWED:
        raise ValueError('Maximum allowed range of days is {}. But {} is configured.'.format(MAX_DAYS_RANGE_ALLOWED, (end - start).days))

    return start, end


def date_range(start_date: date, end_date: date):
    """
    Iterate through dates.
    """
    if end_date < start_date:
        raise ValueError(f'{start_date} is after the {end_date}')
    for n in range(int((end_date - start_date).days) + 1):
        yield start_date + timedelta(n)


def get_week_borders(any_time: datetime) -> Tuple[date, date]:
    """
    Returns the start end the end of the week.
    """
    start: datetime = any_time - timedelta(days=any_time.weekday())
    return start.date(), start + timedelta(days=6)



from typing import Union


class StringUtils:
    @staticmethod
    def is_blank(value: str) -> bool:
        return not (value and value.strip())

    @staticmethod
    def is_not_blank(value: str) -> bool:
        return bool(value and value.strip())

    @staticmethod
    def strip_lines(input_lines: list[str], strip_lines: bool) -> list[str]:
        result: list[str] = []
        for line in input_lines:
            for row in line.split('\n'):
                if strip_lines:
                    row = row.strip()
                result.append(row)
        return result

    @staticmethod
    def count_max_text_width(text: Union[str, list[str]], left_strip_lines: bool = False) -> int:
        if isinstance(text, str):
            return max([len(StringUtils.strip(line, left_strip_lines, True)) for line in text.split('\n')])
        if isinstance(text, list):
            return max([StringUtils.count_max_text_width(line, left_strip_lines) for line in text])
        raise ValueError(f'Wrong input type. Required: Union[str, list[str]]. Actual: {type(text)}')

    @staticmethod
    def strip(text: str, left_strip: bool, right_strip: bool) -> str:
        if left_strip:
            text = text.lstrip()
        if right_strip:
            text = text.rstrip()
        return text



import logging
import textwrap
from collections import defaultdict
from datetime import date, time, datetime, timedelta
from typing import List, Tuple, Optional, NoReturn

from PIL import Image, ImageDraw
from PIL.ImageFont import FreeTypeFont

from calendar_view.config import i18n, style
from calendar_view.core import data, time_utils
from calendar_view.core.config import CalendarConfig, VerticalAlign
from calendar_view.core.event import Event
from calendar_view.core.round_rectangle import draw_rounded_rectangle
from calendar_view.core.utils import StringUtils


class MultilineTextMetadata(object):
    """
    The required information to draw the text (title or notes) for the event.
    """
    def __init__(self, text: Optional[str] = None, size: tuple[int, int] = (0, 0)):
        self.text: Optional[str] = text
        self.size: tuple[int, int] = size
        self.visible: bool = text is not None and size[0] > 0 and size[1] > 0
        # self.trimmed: bool = False

    def __repr__(self) -> str:
        return f'MultilineTextMetadata[visible: {self.visible}, size: {self.size}, text: {self.text}]'


class CalendarEvents(object):
    def __init__(self, config: CalendarConfig):
        self.config = config
        self.event_image: Image = None
        self.event_draw: ImageDraw = None
        self.full_image: Image = None
        self.events: List[Event] = []

    def draw_grid(self, size: Tuple[float, float]):
        self.event_image = Image.new("RGBA", size, (0, 0, 0, 0))
        self.event_draw = ImageDraw.Draw(self.event_image)

    def add_event(self, event: Event) -> None:
        """
        Skip the empty events with a duration of fewer than 0 seconds.
        Splits events, if needed, to the separate days. The event in the result list has to be for 1 day only.
        Cut the event's time out of the visible time range.
        Validate events.
        """
        if event.get_duration_seconds(self.config) < 1:
            logging.warning(f"Skipping event, the duration is too small: {event}")
            return
        end_date: date = event.get_end_date(self.config)
        start_date: date = event.get_start_date(self.config)
        if end_date < self.config.get_date_range()[0]:
            logging.warning(f"Skipping event, it ends before the visible range: {event}")
            return
        if start_date > self.config.get_date_range()[1]:
            logging.warning(f"Skipping event, it starts after the visible range: {event}")
            return

        if start_date == end_date or ((end_date - start_date).days == 1 and event.end_time == time(0, 0)):
            self.__do_add_event(event)
        else:
            logging.debug(f'Splitting the event: {event}')
            iter_from: date = max(start_date, self.config.get_date_range()[0])
            iter_to: date = min(end_date, self.config.get_date_range()[1])
            for single_date in time_utils.date_range(iter_from, iter_to):
                next_date: date = single_date + timedelta(days=1)
                if single_date == start_date:
                    fr: datetime = datetime.combine(single_date, event.start_time)
                    to: datetime = datetime.combine(next_date, time(0, 0))
                elif single_date == end_date:
                    fr: datetime = datetime.combine(single_date, time(0, 0))
                    to: datetime = datetime.combine(single_date, event.end_time)
                else:
                    fr: datetime = datetime.combine(single_date, time(0, 0))
                    to: datetime = datetime.combine(next_date, time(0, 0))
                self.__do_add_event(Event(title=event.title, style=event.style, start=fr, end=to))

    def __do_add_event(self, event: Event) -> None:
        data.validate_event(event, self.config)
        self.events.append(event)
        logging.debug(f'Added internal event: {event}')

        # if legend is needed
        if self.config.legend is None and event.title is not None:
            y = self.__get_event_y(event.start_time, event.end_time)
            height = y[1] - y[0]
            width = style.day_width
            text_size = style.event_title_font.getsize_multiline(event.title)
            if width < text_size[0] or height < text_size[1]:
                self.config.legend = True

    def group_cascade_events(self) -> NoReturn:
        group_counter: int = 1
        groups: dict[int, list[Event]] = defaultdict(list)
        for i in self.events:
            for j in self.events:
                if j == i:
                    pass
                else:
                    if i.start_time < j.start_time < i.end_time \
                            and i.get_start_date(self.config) == j.get_start_date(self.config):
                        if i.cascade_group == 0 and j.cascade_group == 0:
                            i.cascade_group = group_counter
                            j.cascade_group = group_counter
                            group_counter += 1
                        elif i.cascade_group == 0 and j.cascade_group != 0 or i.cascade_group != 0 \
                                and j.cascade_group == 0:
                            group_index: int = max(i.cascade_group, j.cascade_group)
                            i.cascade_group, j.cascade_group = group_index, group_index
                        elif i.cascade_group != j.cascade_group:
                            for k in self.events:
                                if k.cascade_group == i.cascade_group:
                                    k.cascade_group = j.cascade_group

        for i in self.events:
            if i.cascade_group > 0:
                group: list[Event] = groups[i.cascade_group]
                group.append(i)
        for i in groups.values():
            total_group_cascade: int = len(i)
            event_index: int = 1
            for j in i:
                j.cascade_index = event_index
                event_index += 1
                j.cascade_total = total_group_cascade

        for i in groups.values():
            for j in i:
                for k in i:
                    if j.cascade_index > k.cascade_index and j.start_time < k.start_time:
                        j.cascade_index, k.cascade_index = k.cascade_index, j.cascade_index

    def _draw_event(self, event: Event) -> NoReturn:
        """
        The events have already been split to the separate days. The event is for 1 day only.
        """
        day_number = (event.get_start_date(self.config) - self.config.get_date_range()[0]).days
        x = self.__get_event_x(day_number)
        y = self.__get_event_y(event.start_time, event.end_time)
        event_width: int = x[1] - x[0] - style.line_day_width
        cascade_event_width: int = event_width / event.cascade_total
        x1 = x[0] + style.line_day_width / 2 + (event.cascade_index - 1) * cascade_event_width
        x2 = x1 + cascade_event_width
        p1 = (x1, y[0])
        p2 = (x2, y[1])
        draw_rounded_rectangle(self.event_draw, [p1, p2], style.event_radius, outline=event.style.event_border,
                               fill=event.style.event_fill, width=style.event_border_width)

        if self.config.legend:
            return  # The title and notes are printed in the legend. Skip drawing here.

        cell_inner_size: tuple[int, int] = EventDrawHelper.count_cell_inner_size(x, y)
        if cell_inner_size[0] == 0 or cell_inner_size[1] == 0:
            return  # not possible to draw nothing inside the event cell

        # calculate text block sizes
        title_metadata: MultilineTextMetadata = EventDrawHelper.build_title_metadata(event.title, cell_inner_size)
        notes_inner_size: tuple[int, int] = (
            cell_inner_size[0],
            cell_inner_size[1] - (title_metadata.size[1] + style.event_title_margin if title_metadata.visible else 0)
        )
        notes_metadata: MultilineTextMetadata = EventDrawHelper.build_notes_metadata(event.notes, notes_inner_size)

        total_height: int = EventDrawHelper.count_final_text_height(title_metadata, notes_metadata)
        y_top_offset: int = y[0] + style.event_padding
        # print title
        if title_metadata.visible:
            # calculate the top position of the title multiline text block
            y_text_offset: int = EventDrawHelper.calculate_text_y_position_offset(self.config.title_vertical_align,
                                                                                  box_height=cell_inner_size[1],
                                                                                  text_height=title_metadata.size[1],
                                                                                  total_text_height=total_height)
            # the top-left position of the title block
            title_pos: tuple[int, int] = (
                (p1[0] + p2[0]) / 2 - title_metadata.size[0] / 2,
                y_top_offset + y_text_offset
            )
            self.event_draw.multiline_text(title_pos, title_metadata.text, align='center',
                                           font=style.event_title_font, fill=style.event_title_color)
            # update offset for notes
            y_top_offset = title_pos[1] + title_metadata.size[1] + style.event_title_margin

        # print notes
        if notes_metadata.visible:
            y_text_offset: int = 0  # draw the notes right after the title
            if not title_metadata.visible:
                # if the title is not visible, calculate the top-left position of the notes multiline text block
                y_text_offset = EventDrawHelper.calculate_text_y_position_offset(self.config.title_vertical_align,
                                                                                 box_height=cell_inner_size[1],
                                                                                 text_height=notes_metadata.size[1],
                                                                                 total_text_height=total_height)
            # the top-left position of the notes block
            notes_pos: tuple[int, int] = (
                p1[0] + style.event_padding,
                y_top_offset + y_text_offset
            )
            self.event_draw.multiline_text(notes_pos, notes_metadata.text, align='left',
                                           font=style.event_notes_font, fill=style.event_notes_color)

    def destroy(self):
        del self.event_image
        del self.event_draw
        del self.full_image

    def draw_events(self) -> Image:
        for e in self.events:
            self._draw_event(e)
        return self.event_image

    def draw_legend(self) -> Image:
        if not self.config.legend or len(self.events) == 0:
            return None
        width = 0
        height = 0
        for e in self.events:
            text = self._get_event_legend_text(e)
            text_width, text_height = style.legend_name_font.getsize_multiline(text)
            width = max(width, text_width)
            height += text_height

        width += style.legend_padding_left + style.legend_padding_right
        height += (len(self.events) - 1) * style.legend_spacing + style.legend_padding_top + style.legend_padding_bottom

        legend_image: Image = Image.new("RGBA", (width, height), (0, 0, 0, 0))
        legend_draw = ImageDraw.Draw(legend_image)

        x = style.legend_padding_left
        y = style.legend_padding_top
        for e in self.events:
            _, text_height = style.event_title_font.getsize_multiline(e.title)
            text = self._get_event_legend_text(e)
            legend_draw.multiline_text((x, y), text, font=style.legend_name_font, fill=style.legend_name_color)
            y += text_height + style.legend_spacing

        del legend_draw
        return legend_image

    def _get_event_legend_text(self, event: Event) -> str:
        date_text = self._get_day_title(event.get_start_date(self.config))
        time_text = '{:%H:%M} - {:%H:%M}'.format(event.start_time, event.end_time)
        return '{}, {} - {}'.format(date_text, time_text, event.title)

    def _get_day_title(self, day: date) -> str:
        weekday = i18n.day_of_week(day.weekday(), self.config.lang)
        if not self.config.show_date:
            return weekday
        else:
            date = day.strftime('%d.%m') + (day.strftime('.%Y') if self.config.show_year else '')
            return '{}, {}'.format(weekday, date)

    def __get_event_y(self, start: time, end: time):
        start_hour: int = start.hour
        end_hour: int = 24 if (end.hour == 0 and end.minute == 0) else end.hour
        config_start_hour = self.config.get_hours_range()[0]
        if config_start_hour > 0:
            start_hour -= config_start_hour
            end_hour -= config_start_hour

        y_start = style.padding_vertical + style.hour_height + start_hour * style.hour_height + (
                    start.minute / 60) * style.hour_height
        y_end = style.padding_vertical + style.hour_height + end_hour * style.hour_height + (
                    end.minute / 60) * style.hour_height
        return y_start, y_end

    @staticmethod
    def __get_event_x(day_number: int):
        x_start = style.padding_horizontal + day_number * style.day_width
        return x_start, x_start + style.day_width


class EventDrawHelper:
    @staticmethod
    def count_cell_inner_size(x: tuple[float, float], y: tuple[float, float]) -> tuple[int, int]:
        return (
            max(0, int(x[1] - x[0] - 2 * style.line_day_width) - 2 * style.event_padding),
            max(0, int(y[1] - y[0] - 2 * style.line_day_width) - 2 * style.event_padding)
        )

    @staticmethod
    def count_final_text_height(title: MultilineTextMetadata, notes: MultilineTextMetadata) -> int:
        """
        Count the final height for the blocks title and notes in hte calendar event
        """
        height: int = 0
        if title.visible:
            height += title.size[1]
        if notes.visible:
            height += notes.size[1]
        return height

    @staticmethod
    def build_title_metadata(title: Optional[str], cell_inner_size: tuple[int, int]) -> MultilineTextMetadata:
        """
        Try to fit the title in the event inner cell. Split the text into the multiple lines if required.
        """
        return EventDrawHelper.__build_text_metadata(title, cell_inner_size, style.event_title_font, True)

    @staticmethod
    def build_notes_metadata(notes: Optional[str], notes_inner_size: tuple[int, int]) -> MultilineTextMetadata:
        """
        Try to fit notes in the event inner cell. Split the text into the multiple lines if required.
        """
        return EventDrawHelper.__build_text_metadata(notes, notes_inner_size, style.event_notes_font, False)

    @staticmethod
    def calculate_text_y_position_offset(vertical_align: VerticalAlign, box_height: int, text_height: int,
                                         total_text_height: int) -> int:
        """
        Calculates the offset of the text block.
        :param vertical_align: the alignment of the text block
        :param box_height: the height of the inner event cell
        :param text_height: the height of the title or notes multiline text block
        :param total_text_height: the height of the title+notes text blocks
        """
        if vertical_align == 'top':
            return 0
        if vertical_align == 'center':
            return int(box_height / 2 - text_height / 2)
        if vertical_align == 'bottom':
            return max(0, box_height - total_text_height)
        raise RuntimeError(f'Wrong vertical align value: {vertical_align}')

    @staticmethod
    def __build_text_metadata(text: Optional[str], box_size: tuple[int, int], font: FreeTypeFont, strip_lines: bool) \
            -> MultilineTextMetadata:
        """
        Try to fit text in the given box. Split the text into the multiple lines if required.
        """
        if not text or len(text.strip()) == 0:
            return MultilineTextMetadata()
        if box_size[0] <= 0 or box_size[1] <= 0:
            return MultilineTextMetadata()

        text = text.strip()
        text_size: tuple[int, int] = font.getsize_multiline(text)
        if EventDrawHelper.__fits_the_borders(text_size, box_size):
            return MultilineTextMetadata(text, text_size)
        elif EventDrawHelper.__fits_the_width(text_size, box_size):
            return MultilineTextMetadata(text, text_size)

        max_width: int = StringUtils.count_max_text_width(text, strip_lines)
        base_new_text_width: int = int(box_size[0] * max_width / text_size[0])

        new_text: str = ''
        new_text_size: tuple[int, int] = (0, 0)
        for retry_count in range(0, 12, 2):
            new_test_width: int = base_new_text_width - retry_count
            if new_test_width <= 0:
                return MultilineTextMetadata(text, text_size)  # no place to fit the text

            lines: list[str] = textwrap.wrap(text, width=new_test_width, replace_whitespace=False)
            lines = StringUtils.strip_lines(lines, strip_lines)
            new_text = '\n'.join(lines)
            new_text_size = font.getsize_multiline(new_text)
            if EventDrawHelper.__fits_the_borders(new_text_size, box_size):
                return MultilineTextMetadata(new_text, new_text_size)
            elif EventDrawHelper.__fits_the_width(new_text_size, box_size):
                return MultilineTextMetadata(new_text, new_text_size)
        raise RuntimeError('Not possible to wrap the text to fit the width.')
        # textwrap.shorten(title, width=20, placeholder=' ...')

    @staticmethod
    def __fits_the_borders(size: tuple[int, int], border: tuple[int, int]) -> bool:
        return size[0] <= border[0] and size[1] <= border[1]

    @staticmethod
    def __fits_the_width(size: tuple[int, int], border: tuple[int, int]) -> bool:
        return size[0] <= border[0]
